"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const epic_sync_loops_1 = require("epic-sync-loops");
class epicStorage {
    //Constructor
    constructor({ storageName = "storage", rootDir = path_1.default.join(__dirname, "../"), prefix = "epic_", serializer = true, escaper = false, logs = false } = {}) {
        //Defaults
        this.initialized = false;
        this.defaultSchema = "schema";
        this.logDirectory = path_1.default.join(__dirname, "../logs");
        this.logFileName = "epic_logs.log";
        this.schemaList = [];
        this.schema = null;
        this.currentFile = "";
        this.writing = false;
        this.unescapedKey = false;
        this.init = () => {
            return new Promise((resolve) => {
                if (!this.initialized)
                    this.resolveDir().then(() => {
                        this.initialized = true;
                        fs_1.default.readdir(this.storageDir, (error, files) => {
                            if (!error)
                                files.forEach(file => {
                                    if (!this.hasSchema(file.replace(".json", "")))
                                        this.schemaList.push(file.replace(".json", ""));
                                });
                        });
                        this.addSchema(this.defaultSchema, false).then(() => {
                            this.addItems({
                                lastAccessed: new Date(),
                            }, this.defaultSchema, false).then(() => {
                                this.getItem("lastAccessedSchema", this.defaultSchema, false).then((lastAccessedSchema) => {
                                    if (typeof lastAccessedSchema != "undefined" && lastAccessedSchema != this.defaultSchema)
                                        this.useSchema(lastAccessedSchema);
                                    resolve();
                                }).catch(() => {
                                    resolve();
                                });
                            }).catch(error => {
                                resolve();
                            });
                        });
                    });
                else
                    resolve();
            });
        };
        this.resolveDir = () => {
            return new Promise((resolve) => {
                //Create Directory If Not Exists
                fs_1.default.exists(this.storageDir, exists => {
                    if (!exists)
                        fs_1.default.mkdir(this.storageDir, {
                            recursive: true,
                        }, () => {
                            resolve();
                        });
                    else
                        resolve();
                });
            });
        };
        this.escapeHtml = (html) => {
            return (this.escaper) ? html
                .replace(/&/g, "&amp;")
                .replace(/</g, "&lt;")
                .replace(/>/g, "&gt;")
                .replace(/"/g, "&quot;")
                .replace(/'/g, "&#039;") : html;
        };
        this.unescapeHtml = (html) => {
            return (this.escaper) ? html
                .replace(/&amp;/g, "&")
                .replace(/&lt;/g, "<")
                .replace(/&gt;/g, ">")
                .replace(/&quot;/g, '"')
                .replace(/&#039;/g, "'") : html;
        };
        this.replacer = (key, value) => {
            if (typeof value === 'function')
                return value.toString();
            return value;
        };
        this.reviver = (key, value) => {
            if (typeof value === 'string' && value.indexOf('function ') === 0) {
                let functionTemplate = `(${value})`;
                return eval(functionTemplate);
            }
            else
                return value;
        };
        this.serialize = (data, space, normal = false) => {
            return JSON.stringify(data, ((!normal) ? this.replacer : undefined), space);
        };
        this.deSerialize = (data, normal = false) => {
            return JSON.parse(data, ((!normal) ? this.reviver : undefined));
        };
        this.addSchema = (schemaName, use = true) => {
            //Create Schema If Not Exists
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                let schema = path_1.default.join(this.storageDir, schemaName + ".json");
                if (!fs_1.default.existsSync(schema))
                    fs_1.default.writeFile(schema, "{}", () => {
                        if (!this.hasSchema(schemaName)) {
                            this.schemaList.push(schemaName);
                        }
                        if (this.schema == null && use)
                            this.useSchema(schemaName);
                        this.logEvent({
                            event: "epicStorage.schema.added",
                            reference: schemaName
                        });
                        resolve();
                    });
                else {
                    if (!this.hasSchema(schemaName)) {
                        this.schemaList.push(schemaName);
                    }
                    if (this.schema == null && use)
                        this.useSchema(schemaName);
                    resolve();
                }
            });
        };
        this.hasSchema = (schemaName) => {
            if (this.listSchema().includes(schemaName))
                return true;
            else
                return false;
        };
        this.listSchema = () => {
            return this.schemaList.filter((v, i, self) => {
                return self.indexOf(v) === i;
            });
        };
        this.useSchema = (schemaName) => {
            if (!this.initialized)
                throw new Error("Epic Storage Has Not Been Initialized Yet!");
            if (this.hasSchema(schemaName)) {
                if (this.schema != schemaName) {
                    if (this.schema != null)
                        this.logEvent({
                            event: "epicStorage.schema.changed",
                            reference: schemaName
                        });
                    this.schema = schemaName;
                    this.addItems({
                        lastAccessedSchema: schemaName,
                        lastAccessedSchemaTime: new Date(),
                    }, this.defaultSchema);
                    this.currentFile = path_1.default.join(this.storageDir, this.schema + ".json");
                }
            }
            else
                throw new Error(`We Are Unable To Find The (${schemaName}) Schema!`);
            return this;
        };
        this.usingSchema = () => {
            return this.schema;
        };
        this.removeSchema = (schemaName) => {
            if (!this.initialized)
                throw new Error("Epic Storage Has Not Been Initialized Yet!");
            if (typeof schemaName == "undefined")
                if (this.schema != null) {
                    schemaName = this.schema;
                }
                else
                    throw new Error(`No Schema Has Been Selected!`);
            let schema = path_1.default.join(this.storageDir, schemaName + ".json");
            fs_1.default.unlinkSync(schema);
            if (this.hasSchema(schemaName)) {
                // @ts-ignore
                this.schemaList.remove(schemaName);
                this.logEvent({
                    event: "epicStorage.schema.removed",
                    reference: schemaName
                });
                this.addItems({
                    schemas: this.listSchema(),
                    schemasCount: this.listSchema().length
                }, this.defaultSchema);
            }
            if (this.schema == schemaName)
                this.schema = null;
            return this;
        };
        this.renameSchema = (newSchemaName, oldSchemaName) => {
            if (!this.initialized)
                throw new Error("Epic Storage Has Not Been Initialized Yet!");
            if (typeof oldSchemaName == "undefined")
                if (this.schema != null)
                    oldSchemaName = this.schema;
                else
                    throw new Error(`No Schema Has Been Selected!`);
            if (!this.hasSchema(oldSchemaName))
                throw new Error(`Old Schema Not Found!`);
            else if (newSchemaName == oldSchemaName)
                return this;
            else if (this.hasSchema(newSchemaName))
                throw new Error(`A Same Schema Already Exists!`);
            //Else Rename
            let oldSchema = path_1.default.join(this.storageDir, oldSchemaName + ".json");
            let newSchema = path_1.default.join(this.storageDir, newSchemaName + ".json");
            if (!fs_1.default.existsSync(newSchema))
                fs_1.default.renameSync(oldSchema, newSchema);
            // @ts-ignore
            this.schemaList.remove(oldSchemaName);
            if (!this.hasSchema(newSchemaName))
                this.schemaList.push(newSchemaName);
            if (this.schema == oldSchemaName)
                this.useSchema(newSchemaName);
            this.logEvent({
                event: "epicStorage.schema.renamed",
                reference: {
                    fromSchemaName: oldSchemaName,
                    toSchemaName: newSchemaName
                }
            });
            return this;
        };
        this.clearSchema = (schema, throwError = true) => {
            if (!this.initialized)
                throw new Error("Epic Storage Has Not Been Initialized Yet!");
            let currentFile;
            if (typeof schema == "undefined")
                currentFile = this.currentFile;
            else if (this.hasSchema(schema))
                currentFile = path_1.default.join(this.storageDir, schema + ".json");
            else if (throwError)
                throw new Error(`Schema '${schema}' Not Found!`);
            else
                return this;
            //Clear Data
            if (fs_1.default.existsSync(currentFile))
                fs_1.default.writeFileSync(currentFile, "{}", {
                    flag: 'w',
                });
            this.logEvent({
                event: "epicStorage.schema.cleared",
                reference: ((schema) ? schema : this.schema)
            });
            return this;
        };
        this.addItem = (key, value, schema, throwError = true) => {
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                let loop = new epic_sync_loops_1.epicSyncLoops().loop(() => {
                    if (this.writing) {
                        loop.next();
                    }
                    else {
                        this.writing = true;
                        let currentFile;
                        if (typeof schema == "undefined")
                            if (this.schema == null)
                                if (throwError) {
                                    this.writing = false;
                                    reject(`Please Choose A Schema First!`);
                                    return false;
                                }
                                else {
                                    this.writing = false;
                                    resolve();
                                    return true;
                                }
                            else
                                currentFile = this.currentFile;
                        else if (this.hasSchema(schema))
                            currentFile = path_1.default.join(this.storageDir, schema + ".json");
                        else if (throwError) {
                            this.writing = false;
                            reject(`Schema '${schema}' Not Found!`);
                            return false;
                        }
                        else {
                            this.writing = false;
                            resolve();
                            return true;
                        }
                        //Get Previous Data
                        this.getItems(schema).then((data) => {
                            data[this.escapeHtml(key)] = (typeof value == "string") ? this.escapeHtml(value) : value;
                            fs_1.default.writeFile(currentFile, this.serialize(data, 2, this.serializer), {
                                flag: 'w',
                            }, (error) => {
                                if (error && throwError) {
                                    this.writing = false;
                                    reject(error);
                                    return false;
                                }
                                else if (!error)
                                    this.logEvent({
                                        event: "epicStorage.item.added",
                                        reference: key
                                    });
                                this.writing = false;
                                resolve();
                            });
                        }).catch(error => {
                            this.writing = false;
                            reject(error);
                        });
                        loop.end();
                    }
                });
            });
        };
        this.addItems = (data, schema, throwError = true) => {
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                let loop = (new epic_sync_loops_1.epicSyncLoops).loop((i, v) => {
                    this.addItem(i, v, schema, throwError).then(() => {
                        loop.next();
                    }).catch(error => {
                        if (throwError)
                            reject(error);
                    });
                }, data);
                let _loop = new epic_sync_loops_1.epicSyncLoops().loop(() => {
                    resolve();
                    _loop.end();
                }).after(loop);
            });
        };
        this.getItems = (schema, throwError = true, parse = true) => {
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                let currentSchema = this.schema;
                if (typeof schema != "undefined" && this.hasSchema(schema))
                    currentSchema = schema;
                if (currentSchema != null) {
                    let currentFile = path_1.default.join(this.storageDir, currentSchema + ".json");
                    let data = "{}";
                    if (fs_1.default.existsSync(currentFile)) {
                        //@ts-ignore
                        data = fs_1.default.readFileSync(currentFile);
                        try {
                            if (parse)
                                resolve(this.deSerialize(data, this.serializer));
                            else
                                resolve(data);
                        }
                        catch (ex) {
                            resolve({});
                        }
                    }
                    else {
                        resolve({});
                    }
                    return true;
                }
                else if (throwError) {
                    reject(`No Schema Has Been Selected!`);
                    return false;
                }
                else
                    resolve({});
                return true;
            });
        };
        this.getAllItems = (schema, throwError = true, parse = true) => {
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                this.getItems(schema, throwError, parse).then((data) => {
                    if (typeof data == "string") {
                        reject(`An Error Has Been Occured! System Was Unable To Parse The Data!`);
                        return false;
                    }
                    else {
                        let unescapedData = {};
                        let loop = new epic_sync_loops_1.epicSyncLoops().loop((i, v) => {
                            if (this.unescapedKey)
                                unescapedData[i] = (typeof v == "string") ? this.unescapeHtml(v) : v;
                            else
                                unescapedData[this.unescapeHtml(i)] = (typeof v == "string") ? this.unescapeHtml(v) : v;
                            loop.next();
                        }, data);
                        let _loop = new epic_sync_loops_1.epicSyncLoops().loop(() => {
                            resolve(unescapedData);
                            _loop.end();
                        }).after(loop);
                    }
                }).catch(reject);
            });
        };
        this.getItem = (key, schema, throwError = true, parse = true) => {
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                this.unescapedKey = true;
                this.getAllItems(schema, throwError, parse).then(data => {
                    // @ts-ignore
                    if (typeof data[this.escapeHtml(key)] == "undefined")
                        if (throwError)
                            reject(`Results Are Empty (May Be Invalid Data ID '${this.escapeHtml(key)}')!`);
                        else
                            resolve(undefined);
                    else {
                        this.logEvent({
                            event: "epicStorage.item.retrived",
                            reference: key
                        });
                        // @ts-ignore
                        resolve(data[this.escapeHtml(key)]);
                    }
                    this.unescapedKey = false;
                }).catch(error => {
                    reject(error);
                });
            });
        };
        this.removeItem = (key, schema, throwError = true) => {
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                let currentFile;
                if (typeof schema == "undefined")
                    currentFile = this.currentFile;
                else if (this.hasSchema(schema))
                    currentFile = path_1.default.join(this.storageDir, schema + ".json");
                else if (throwError) {
                    reject(`Schema '${schema}' Not Found!`);
                    return false;
                }
                else {
                    resolve();
                    return true;
                }
                this.getItems(schema).then((data) => {
                    delete data[this.escapeHtml(key)];
                    fs_1.default.writeFile(currentFile, this.serialize(data, 2, this.serializer), {
                        flag: 'w',
                    }, (error) => {
                        if (error && throwError) {
                            reject(error);
                            return false;
                        }
                        else if (!error)
                            this.logEvent({
                                event: "epicStorage.item.removed",
                                reference: key
                            });
                        resolve();
                    });
                });
            });
        };
        this.hasItem = (key, schema) => {
            return new Promise((resolve, reject) => {
                this.getItem(key, schema).then(resolve, reject);
            });
        };
        this.removeItems = (keys, schema, throwError = true) => {
            return new Promise((resolve, reject) => {
                if (!this.initialized) {
                    reject("Epic Storage Has Not Been Initialized Yet!");
                    return false;
                }
                let loop = (new epic_sync_loops_1.epicSyncLoops).loop((i, v) => {
                    this.removeItem(v, schema).then(() => {
                        loop.next();
                    }).catch(error => {
                        if (throwError)
                            reject(error);
                    });
                }, keys);
                let _loop = new epic_sync_loops_1.epicSyncLoops().loop(() => {
                    resolve();
                    _loop.end();
                }).after(loop);
            });
        };
        this.destroy = (logs = false) => {
            fs_1.default.rmdirSync(this.storageDir, { recursive: true });
            if (logs)
                fs_1.default.rmdirSync(this.logDirectory, { recursive: true });
            this.logEvent({
                event: "epicStorage.destroyed",
                reference: this.storageDir
            });
            this.schemaList = [];
            this.schema = null;
            this.currentFile = "";
            return this;
        };
        this.logConfig = (fileName, directory) => {
            this.logDirectory = directory;
            this.logFileName = fileName;
            return this;
        };
        this.clearLog = () => {
            fs_1.default.rmdirSync(this.logDirectory, { recursive: true });
            return this;
        };
        this.logEvent = (data, force = false) => {
            if (this.logs || force) {
                // @ts-ignore
                data.time = new Date();
                //Create Directory If Not Exists
                if (!fs_1.default.existsSync(this.logDirectory))
                    fs_1.default.mkdirSync(this.logDirectory, {
                        recursive: true,
                    });
                fs_1.default.appendFileSync(path_1.default.join(this.logDirectory, this.logFileName), this.serialize(data, undefined, this.serializer) + ",\n\n");
                //Log
                console.log("Event Logger:", data);
            }
            return this;
        };
        this.storageName = storageName;
        this.rootDir = rootDir.replace(/^\/|\/$/g, '') + "/";
        this.prefix = prefix;
        this.serializer = serializer;
        this.escaper = escaper;
        this.logs = logs;
        this.baseName = this.prefix + this.storageName;
        this.storageDir = path_1.default.join(this.rootDir, this.baseName);
        // @ts-ignore Add Feature To Array Constructor
        Array.prototype.remove = function () {
            var what, a = arguments, L = a.length, ax;
            while (L && this.length) {
                what = a[--L];
                while ((ax = this.indexOf(what)) !== -1) {
                    this.splice(ax, 1);
                }
            }
            return this;
        };
        return this;
    }
}
exports.epicStorage = epicStorage;
;
//# sourceMappingURL=epicStorage.js.map